import time
from contextlib import nullcontext
from typing_extensions import Annotated
from rich.console import Console
from rich.table import Table
from rich.markdown import Markdown
import typer
from cpr_sdk.search_adaptors import VespaSearchAdapter
from cpr_sdk.models.search import SearchParameters, Passage, Document, SearchResponse
from cpr_sdk.vespa import build_vespa_request_body, parse_vespa_response
from cpr_sdk.config import VESPA_URL

SCORES_NUM_DECIMALS = 3


def get_rank_feature_names(search_response: SearchResponse) -> list[str]:
    """
    Get names of rank features from a search response.

    Rank features surface the scores given to individual parts of the query, and are
    defined in the Vespa schema.
    """
    rank_feature_names = set()
    for family in search_response.families:
        for hit in family.hits:
            if hit.rank_features:
                rank_feature_names.update(
                    k for k in hit.rank_features.keys() if not k.startswith("vespa")
                )
    rank_feature_names = sorted(list(rank_feature_names))

    return rank_feature_names


def add_tokens_summary_to_yql(yql: str) -> str:
    """Amend the summary requested in a YQL query to return tokens."""

    return yql.replace("summary(search_summary)", "summary(search_summary_with_tokens)")


def main(
    query: str = typer.Argument(..., help="The search query to run."),
    exact_match: bool = False,
    limit: int = 20,
    show_rank_features: bool = False,
    page_results: Annotated[
        bool,
        typer.Option(
            help="Whether to use the default terminal pager to show results. Disable with `--no-page-results` if you want to redirect the output to a file."
        ),
    ] = True,
    experimental_tokens: Annotated[
        bool,
        typer.Option(
            help="Whether to include tokens in the summary. Tokens are not in the final Vespa response model, so this requires setting a breakpoint on the raw response."
        ),
    ] = False,
):
    """Run a search query with different rank profiles."""
    console = Console()
    search_adapter = VespaSearchAdapter(VESPA_URL)
    search_parameters = SearchParameters(
        query_string=query, exact_match=exact_match, limit=limit
    )
    request_body = build_vespa_request_body(search_parameters)

    if experimental_tokens:
        print(
            "WARNING: tokens are not fed into the final Vespa response, so you will see no change unless you set a breakpoint just after `search_response_raw` following these lines."
        )
        request_body["yql"] = add_tokens_summary_to_yql(request_body["yql"])

    start_time = time.time()
    search_response_raw = search_adapter.client.query(body=request_body)
    request_time = time.time() - start_time

    # Debugging steps for showing tokens
    # from rich import print as rprint
    # rprint(search_response_raw.json)
    # breakpoint()

    search_response = parse_vespa_response(search_response_raw)
    n_results = len(search_response.families)
    rank_feature_names = get_rank_feature_names(search_response)

    pager = console.pager(styles=True, links=True) if page_results else nullcontext()

    with pager:
        console.print(Markdown("# Query"))
        console.print(f"Text: {query}")
        console.print(f"Exact match: {exact_match}")
        console.print(f"Limit: {limit}")
        console.print(f"Request time: {request_time:.3f}s")
        console.print("Request body:")
        console.print_json(data=request_body)

        console.print(Markdown("# Families"))
        table = Table(show_header=True, header_style="bold", show_lines=True)
        table.add_column("Family Name")
        table.add_column("Geography")
        table.add_column("Score")
        table.add_column("Hits")
        table.add_column("Slug")

        for family in search_response.families:
            family_data = family.hits[0].model_dump()
            table.add_row(
                family_data["family_name"],
                family_data["family_geography"],
                str(round(family_data["relevance"], SCORES_NUM_DECIMALS)),
                str(len(family.hits)),
                family_data["family_slug"],
            )

        console.print(table)

        console.print(Markdown("# Results"))

        for idx, family in enumerate(search_response.families, start=1):
            family_data = family.hits[0].model_dump()
            console.rule(
                title=f"Family {idx}/{n_results}: '{family_data['family_name']}' ({family_data['family_geography']}). Score: {round(family_data['relevance'], 3)}"
            )
            family_url = f"https://app.climatepolicyradar.org/document/{family_data['family_slug']}"
            details = f"""
            [bold]Total hits:[/bold] {len(family.hits)}
            [bold]Family:[/bold] [link={family_url}]{family_data['family_import_id']}[/link]
            [bold]Family slug:[/bold] {family_data['family_slug']}
            [bold]Geography:[/bold] {family_data['family_geography']}
            [bold]Relevance:[/bold] {family_data['relevance']}
            """

            console.print(details)
            console.print(
                f"[bold]Description:[/bold] {family_data['family_description']}"
            )
            console.print("\n[bold]Hits:[/bold]")

            # Create table headers
            table = Table(show_header=True, header_style="bold", show_lines=True)
            table.add_column("Text")
            table.add_column("Score")
            table.add_column("Type")
            table.add_column("TB ID")
            table.add_column("Doc ID")
            if show_rank_features:
                for feature_name in rank_feature_names:
                    table.add_column(feature_name)

            for hit in family.hits:
                if isinstance(hit, Passage):
                    hit_type = "Text block"
                    text = hit.text_block
                    tb_id = hit.text_block_id
                    doc_id = hit.document_import_id
                elif isinstance(hit, Document):
                    hit_type = "Document"
                    text = "<see family description>"
                    tb_id = "-"
                    doc_id = hit.document_import_id
                else:
                    raise ValueError(f"Whoops! Unknown hit type {type(hit)}")

                rank_feature_values = (
                    [hit.rank_features.get(name) for name in rank_feature_names]
                    if (show_rank_features and hit.rank_features is not None)
                    else []
                )
                rank_feature_values = [
                    str(round(v, SCORES_NUM_DECIMALS)) if v is not None else "-"
                    for v in rank_feature_values
                ]

                table.add_row(
                    text,
                    str(round(hit.relevance, SCORES_NUM_DECIMALS)),
                    hit_type,
                    tb_id,
                    doc_id,
                    *rank_feature_values,
                )

            console.print(table)


if __name__ == "__main__":
    typer.run(main)



import os
from typing import Optional
from dotenv import load_dotenv


load_dotenv()

VESPA_URL: Optional[str] = os.environ.get("VESPA_URL")
if VESPA_URL is not None:
    VESPA_URL = VESPA_URL.rstrip("/")



from typing import Optional, Any, Callable, Sequence, Literal

import pytest
from pydantic import BaseModel, Field, model_validator

from cpr_sdk.models.search import filter_fields, Filters


class TestCase(BaseModel):
    """A model to test the search functionality. Used by the `executors` module."""

    id: Optional[str] = Field(
        description="A unique identifier for the test case", default=None
    )
    search_terms: str = Field(description="The terms which will be searched for.")
    filters: Optional[dict[str, Sequence[str]]] = Field(
        description=("Filters to be applied to the search terms."), default=None
    )
    document_id: Optional[str] = Field(
        description=("An optional Document ID to filter the search on."), default=None
    )
    exact_match: bool = Field(
        description=("Whether to run an exact match search."),
        default=False,
    )
    description: Optional[str] = Field(
        description=(
            "An optional description of the test case. Can be used to explain "
            "what the test intention is, any assumptions made, links to "
            "discussions which motivated the test, etc."
        ),
        default=None,
    )
    known_failure: bool = Field(
        description=(
            "If True, the test is expected to fail. If False, the test is "
            "expected to pass. NOTE: this should be set to False for every test before starting to evaluate search. Setting this to True is a way of zooming in on failures to fix in a round of fixing. "
        ),
        default=False,
    )

    def __init__(self, **data):
        if "id" not in data or data["id"] is None:
            data["id"] = "__".join(
                [
                    str(self),
                    data.get("description", ""),
                    data["search_terms"],
                ]
            )
        super().__init__(**data)

    @property
    def param(self):
        """Return the test case as a pytest parameter"""
        return pytest.param(
            self,
            id=self.id,
            marks=[pytest.mark.xfail(strict=True)] if self.known_failure else [],
        )

    @model_validator(mode="after")
    def check_filter_fields(self):
        """Check that the filter fields are valid."""
        if self.filters is not None:
            supplied_filter_fields = set(self.filters.keys())
            if not supplied_filter_fields.issubset(set(filter_fields)):
                invalid_filter_fields = supplied_filter_fields - set(filter_fields)
                raise ValueError(
                    f"Filter fields {invalid_filter_fields} are not valid."
                )
        return self

    def get_search_filters(self) -> Optional[Filters]:
        """Return the search filters."""
        if self.filters is None:
            return None
        filters_mapped_names = {
            filter_fields[field]: values for field, values in self.filters.items()
        }
        return Filters.model_validate(filters_mapped_names)


class TopFamiliesTestCase(TestCase):
    """Dictates which should be the first (or top) families for a given search."""

    expected_family_slugs: list[str] = Field(
        description="The expected family slugs for the top results."
    )
    strict_order: bool = Field(
        description=(
            "Whether the expected family slugs should be in the exact order "
            "specified."
        ),
        default=False,
    )

    def __str__(self) -> str:  # noqa: D105
        return "TopFamiliesTestCase"

    @model_validator(mode="after")
    def check_expected_family_slugs_unique(self):
        """Check that the expected family slugs are unique."""
        if len(self.expected_family_slugs) != len(set(self.expected_family_slugs)):
            raise ValueError("expected_family_slugs must be unique")
        return self


class FieldCharacteristicsTestCase(TestCase):
    """Dictates the characteristics which any or all of the top k results should have for a given search."""

    test_field: Literal["family_name", "text_block_text", "geographies"] = Field(
        description="The field to test for the expected characteristics."
    )
    characteristics_test: Callable[[Any], bool] = Field(
        description="A function which returns True if the expected characteristics for the field value are met, and False otherwise."
    )
    k: int = Field(
        description="The number of families to check for the expected characteristics.",
    )
    all_or_any: Literal["all", "any"] = Field(
        description="Whether all or any of the words in the search terms should be in the field value.",
        default="all",
    )

    def __str__(self) -> str:  # noqa: D105
        return "CharacteristicsTestCase"


class FamiliesInTopKTestCase(TestCase):
    """Dictates families which should be anywhere within the top k results for a given search."""

    expected_family_slugs: list[str] = Field(
        description="Family slugs which should appear in the top k results."
    )
    forbidden_family_slugs: Optional[list[str]] = Field(
        description="Family slugs which should not appear in the top k results.",
        default=None,
    )
    k: int = Field(
        description="The number of results to check for the expected family slugs. Defaults to 20: the number of results shown on the first page of the CPR tools.",
        default=20,
    )

    @model_validator(mode="after")
    def check_expected_family_slugs_unique(self):
        """Check that the expected family slugs are unique."""
        if len(self.expected_family_slugs) != len(set(self.expected_family_slugs)):
            raise ValueError("expected_family_slugs must be unique")
        return self

    @model_validator(mode="after")
    def check_forbidden_family_slugs_unique(self):
        """Check that the forbidden family slugs are unique."""
        if self.forbidden_family_slugs is not None and len(
            self.forbidden_family_slugs
        ) != len(set(self.forbidden_family_slugs)):
            raise ValueError("expected_family_slugs must be unique")
        return self

    def __str__(self) -> str:  # noqa: D105
        return "TopCharacteristicsTestCase"


class SearchComparisonTestCase(TestCase):
    """
    Compare two searches to each other. Compares the top k results.

    For a whole-database search families are compared; for a single document search text of text blocks.
    """

    search_terms_to_compare: str = Field(
        description="The terms to compare the search_terms to."
    )

    k: int = Field(
        description="The number of results to compare for the expected family slugs. Defaults to 20: the number of results shown on the first page of the CPR tools.",
        default=20,
    )

    minimum_families_overlap: float = Field(
        description="The desired proportion of the top k results which have the same family slug.",
        strict=True,
        gt=0,
        le=1,
    )

    strict_order: bool = Field(
        description=(
            "Whether the expected family slugs should be in the exact order "
            "specified."
        ),
        default=False,
    )

    @model_validator(mode="after")
    def check_comparison_terms(self):
        """Check that the comparison terms are different from the search terms."""
        if self.search_terms == self.search_terms_to_compare:
            raise ValueError(
                "search_terms and search_terms_to_compare must be different"
            )
        return self

    def __str__(self) -> str:  # noqa: D105
        return "SearchComparisonTestCase"



from cpr_sdk.search_adaptors import VespaSearchAdapter
from cpr_sdk.models.search import SearchParameters, SearchResponse, Passage

from cpr_sdk.search_intention_testing.models import (
    TestCase,
    TopFamiliesTestCase,
    FieldCharacteristicsTestCase,
    FamiliesInTopKTestCase,
    SearchComparisonTestCase,
)


def get_search_response(
    test_case: TestCase,
    limit: int,
    instance_url: str,
) -> SearchResponse:
    """Get a response from a Vespa instance given a test case and a limit."""
    search_adapter = VespaSearchAdapter(instance_url)

    search_parameters = SearchParameters(
        query_string=test_case.search_terms,
        exact_match=test_case.exact_match,
        document_ids=[test_case.document_id] if test_case.document_id else None,
        filters=test_case.get_search_filters(),
        limit=limit,
    )

    return search_adapter.search(search_parameters)


def do_test_top_families(
    test_case: TopFamiliesTestCase,
    instance_url: str,
):
    search_response = get_search_response(test_case, 20, instance_url)

    top_family_slugs = [
        family.hits[0].family_slug
        for family in search_response.families[: len(test_case.expected_family_slugs)]
    ]

    if test_case.strict_order:
        assert top_family_slugs == test_case.expected_family_slugs
    else:
        assert set(top_family_slugs) == set(test_case.expected_family_slugs)


def do_test_families_in_top_k(
    test_case: FamiliesInTopKTestCase,
    instance_url: str,
):
    search_response = get_search_response(test_case, test_case.k, instance_url)

    family_slugs_in_response = [
        family.hits[0].family_slug for family in search_response.families
    ]

    expected_family_slugs_not_in_response = set(
        test_case.expected_family_slugs
    ).difference(set(family_slugs_in_response))

    assert (
        not expected_family_slugs_not_in_response
    ), f"Expected family slugs not found in top {test_case.k} results: {expected_family_slugs_not_in_response}"

    if test_case.forbidden_family_slugs is not None:
        forbidden_slugs_in_response = set(
            test_case.forbidden_family_slugs
        ).intersection(set(family_slugs_in_response))

        assert (
            not forbidden_slugs_in_response
        ), f"Forbidden family slugs found in top {test_case.k} results: {forbidden_slugs_in_response}"


def do_test_field_characteristics(
    test_case: FieldCharacteristicsTestCase,
    instance_url: str,
):
    search_response = get_search_response(test_case, test_case.k, instance_url)

    match test_case.test_field:
        case "family_name":
            field_values = [
                family.hits[0].family_name for family in search_response.families
            ]
        case "text_block_text":
            field_values = [
                hit.text_block
                for family in search_response.families
                for hit in family.hits
                if isinstance(hit, Passage)
            ]  # type: ignore
        case "geographies":
            field_values = [
                family.hits[0].family_geographies for family in search_response.families
            ]
        case _:
            raise ValueError(f"Unknown test field: {test_case.test_field}")

    failing_values = [
        value for value in field_values if not test_case.characteristics_test(value)
    ]

    match test_case.all_or_any:
        case "all":
            assert (
                not failing_values
            ), f"Values of {test_case.test_field} found failing test: {failing_values}"
        case "any":
            passing_values = [
                value for value in field_values if test_case.characteristics_test(value)
            ]
            assert (
                passing_values
            ), f"No values of {test_case.test_field} found passing test. Failing values: {failing_values}."


def do_test_search_comparison(
    test_case: SearchComparisonTestCase,
    instance_url: str,
):
    search_response_a = get_search_response(test_case, test_case.k, instance_url)

    test_case_b = test_case.model_copy(
        update={"search_terms": test_case.search_terms_to_compare}
    )
    search_response_b = get_search_response(test_case_b, test_case.k, instance_url)

    unit_of_comparison = None

    if test_case.document_id:
        results_a = [
            hit.text_block
            for family in search_response_a.families
            for hit in family.hits
            if isinstance(hit, Passage)
        ]
        results_b = [
            hit.text_block
            for family in search_response_b.families
            for hit in family.hits
            if isinstance(hit, Passage)
        ]
        unit_of_comparison = "text blocks"
    else:
        results_a = [
            family.hits[0].family_slug for family in search_response_a.families
        ]
        results_b = [
            family.hits[0].family_slug for family in search_response_b.families
        ]
        unit_of_comparison = "families"

    if test_case.strict_order:
        differences = [
            (idx, result_a, result_b)
            for idx, (result_a, result_b) in enumerate(zip(results_a, results_b))
            if result_a != result_b
        ]

        overlap_proportion = 1 - (
            len(differences) / max(len(results_a), len(results_b))
        )

        assert (
            overlap_proportion >= test_case.minimum_families_overlap
        ), f"Strict overlap between {unit_of_comparison} is less than expected ({overlap_proportion}). Differences: {differences}"
    else:
        if unit_of_comparison == "text blocks":
            # text blocks aren't unique, so we can't use set intersection
            overlap = len([block for block in results_a if block in results_b]) / max(
                len(results_a), len(results_b)
            )
            differences = [block for block in results_a if block not in results_b] + [
                block for block in results_b if block not in results_a
            ]

        else:
            overlap = len(set(results_a).intersection(set(results_b))) / test_case.k
            differences = set(results_a).symmetric_difference(set(results_b))

        assert (
            overlap >= test_case.minimum_families_overlap
        ), f"Overlap between {unit_of_comparison} is less than expected: {overlap} < {test_case.minimum_families_overlap}. Differences: {differences}"
